@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 
 # Hive Flask Quorum
-# Copyright (C) 2008-2015 Hive Solutions Lda.
+# Copyright (c) 2008-2017 Hive Solutions Lda.
 #
 # This file is part of Hive Flask Quorum.
 #
@@ -31,7 +31,7 @@ __revision__ = "$LastChangedRevision$"
 __date__ = "$LastChangedDate$"
 """ The last change date of the module """
 
-__copyright__ = "Copyright (c) 2008-2015 Hive Solutions Lda."
+__copyright__ = "Copyright (c) 2008-2017 Hive Solutions Lda."
 """ The copyright for the module """
 
 __license__ = "Apache License, Version 2.0"
@@ -41,9 +41,12 @@ import json
 import base64
 import random
 import string
+import logging
 
 from . import legacy
+from . import typesf
 from . import config
+from . import common
 from . import exceptions
 
 TIMEOUT = 60
@@ -68,6 +71,16 @@ def try_auth(auth_callback, params, headers = None):
     if headers == None: headers = dict()
     auth_callback(params, headers)
 
+def get_f(*args, **kwargs):
+    name = kwargs.pop("name", "default")
+    kwargs["handle"] = kwargs.get("handle", True)
+    kwargs["redirect"] = kwargs.get("redirect", True)
+    data, response = get_json(*args, **kwargs)
+    info = response.info()
+    mime = info.get("Content-Type", None)
+    file_tuple = (name, mime, data)
+    return typesf.File(file_tuple)
+
 def get(url, auth_callback = None, **kwargs):
     # starts the variable holding the number of
     # retrieves to be used
@@ -80,35 +93,54 @@ def get(url, auth_callback = None, **kwargs):
             if error.code in AUTH_ERRORS and auth_callback:
                 try_auth(auth_callback, kwargs)
             else:
-                raise
+                code = error.getcode()
+                raise exceptions.HTTPDataError(error, code)
 
         # decrements the number of retries and checks if the
         # number of retries has reached the limit
         retries -= 1
         if retries == 0:
-            raise exceptions.HttpError("Data retrieval not possible")
+            raise exceptions.HTTPError("Data retrieval not possible")
 
-def get_json(url, headers = None, handle = None, auth_callback = None, **kwargs):
+def get_json(
+    url,
+    headers = None,
+    handle = None,
+    redirect = None,
+    silent = None,
+    timeout = None,
+    auth_callback = None,
+    **kwargs
+):
     # starts the variable holding the number of
     # retrieves to be used
     retries = 5
 
     while True:
         try:
-            return _get_json(url, handle = handle, headers = headers, **kwargs)
+            return _get_json(
+                url,
+                headers = headers,
+                handle = handle,
+                silent = silent,
+                redirect = redirect,
+                timeout = timeout,
+                **kwargs
+            )
         except legacy.HTTPError as error:
             if error.code in AUTH_ERRORS and auth_callback:
                 try_auth(auth_callback, kwargs)
             else:
                 data_r = error.read()
+                data_r = legacy.str(data_r, encoding = "utf-8")
                 data_s = json.loads(data_r)
-                raise exceptions.JsonError(data_s)
+                raise exceptions.JSONError(data_s)
 
         # decrements the number of retries and checks if the
         # number of retries has reached the limit
         retries -= 1
         if retries == 0:
-            raise exceptions.HttpError("Data retrieval not possible")
+            raise exceptions.HTTPError("Data retrieval not possible")
 
 def post_json(
     url,
@@ -118,6 +150,9 @@ def post_json(
     headers = None,
     mime = None,
     handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
     auth_callback = None,
     **kwargs
 ):
@@ -135,6 +170,9 @@ def post_json(
                 headers = headers,
                 mime = mime,
                 handle = handle,
+                silent = silent,
+                redirect = redirect,
+                timeout = timeout,
                 **kwargs
             )
         except legacy.HTTPError as error:
@@ -142,14 +180,15 @@ def post_json(
                 try_auth(auth_callback, kwargs)
             else:
                 data_r = error.read()
+                data_r = legacy.str(data_r, encoding = "utf-8")
                 data_s = json.loads(data_r)
-                raise exceptions.JsonError(data_s)
+                raise exceptions.JSONError(data_s)
 
         # decrements the number of retries and checks if the
         # number of retries has reached the limit
         retries -= 1
         if retries == 0:
-            raise exceptions.HttpError("Data retrieval not possible")
+            raise exceptions.HTTPError("Data retrieval not possible")
 
 def put_json(
     url,
@@ -159,6 +198,9 @@ def put_json(
     headers = None,
     mime = None,
     handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
     auth_callback = None,
     **kwargs
 ):
@@ -176,6 +218,9 @@ def put_json(
                 headers = headers,
                 mime = mime,
                 handle = handle,
+                silent = silent,
+                redirect = redirect,
+                timeout = timeout,
                 **kwargs
             )
         except legacy.HTTPError as error:
@@ -183,47 +228,83 @@ def put_json(
                 try_auth(auth_callback, kwargs)
             else:
                 data_r = error.read()
+                data_r = legacy.str(data_r, encoding = "utf-8")
                 data_s = json.loads(data_r)
-                raise exceptions.JsonError(data_s)
+                raise exceptions.JSONError(data_s)
 
         # decrements the number of retries and checks if the
         # number of retries has reached the limit
         retries -= 1
         if retries == 0:
-            raise exceptions.HttpError("Data retrieval not possible")
+            raise exceptions.HTTPError("Data retrieval not possible")
 
-def delete_json(url, headers = None, handle = None, auth_callback = None, **kwargs):
+def delete_json(
+    url,
+    headers = None,
+    handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
+    auth_callback = None,
+    **kwargs
+):
     # starts the variable holding the number of
     # retrieves to be used
     retries = 5
 
     while True:
         try:
-            return _delete_json(url, headers = headers, handle = handle, **kwargs)
+            return _delete_json(
+                url,
+                headers = headers,
+                handle = handle,
+                silent = silent,
+                redirect = redirect,
+                timeout = timeout,
+                **kwargs
+            )
         except legacy.HTTPError as error:
             if error.code in AUTH_ERRORS and auth_callback:
                 try_auth(auth_callback, kwargs)
             else:
                 data_r = error.read()
+                data_r = legacy.str(data_r, encoding = "utf-8")
                 data_s = json.loads(data_r)
-                raise exceptions.JsonError(data_s)
+                raise exceptions.JSONError(data_s)
 
         # decrements the number of retries and checks if the
         # number of retries has reached the limit
         retries -= 1
         if retries == 0:
-            raise exceptions.HttpError("Data retrieval not possible")
+            raise exceptions.HTTPError("Data retrieval not possible")
 
 def _get(url, **kwargs):
     values = kwargs or dict()
     data = _urlencode(values)
     url = url + "?" + data if data else url
-    file = _resolve(url, "GET", {}, None, TIMEOUT)
+    file = _resolve(url, "GET", {}, None, False, TIMEOUT)
     contents = file.read()
     return contents
 
-def _get_json(url, headers = None, handle = None, **kwargs):
-    return _method_empty("GET", url, headers = headers, handle = handle, **kwargs)
+def _get_json(
+    url,
+    headers = None,
+    handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
+    **kwargs
+):
+    return _method_empty(
+        "GET",
+        url,
+        headers = headers,
+        handle = handle,
+        silent = silent,
+        redirect = redirect,
+        timeout = timeout,
+        **kwargs
+    )
 
 def _post_json(
     url,
@@ -233,6 +314,9 @@ def _post_json(
     headers = None,
     mime = None,
     handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
     **kwargs
 ):
     return _method_payload(
@@ -244,6 +328,9 @@ def _post_json(
         headers = headers,
         mime = mime,
         handle = handle,
+        silent = silent,
+        redirect = redirect,
+        timeout = timeout,
         **kwargs
     )
 
@@ -255,6 +342,9 @@ def _put_json(
     headers = None,
     mime = None,
     handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
     **kwargs
 ):
     return _method_payload(
@@ -266,25 +356,69 @@ def _put_json(
         headers = headers,
         mime = mime,
         handle = handle,
+        silent = silent,
+        redirect = redirect,
+        timeout = timeout,
         **kwargs
     )
 
-def _delete_json(url, headers = None, handle = None, **kwargs):
-    return _method_empty("DELETE", url, headers = headers, handle = handle, **kwargs)
+def _delete_json(
+    url,
+    headers = None,
+    handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
+    **kwargs
+):
+    return _method_empty(
+        "DELETE",
+        url,
+        headers = headers,
+        handle = handle,
+        silent = silent,
+        redirect = redirect,
+        timeout = timeout,
+        **kwargs
+    )
 
-def _method_empty(name, url, headers = None, handle = None, **kwargs):
+def _method_empty(
+    name,
+    url,
+    headers = None,
+    handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
+    **kwargs
+):
     if handle == None: handle = False
+    if silent == None: silent = False
+    if redirect == None: redirect = False
+    if timeout == None: timeout = TIMEOUT
     values = kwargs or dict()
+    url, scheme, host, authorization, extra = _parse_url(url)
+    if extra: values.update(extra)
     data = _urlencode(values)
-    url, host, authorization = _parse_url(url)
-    headers = headers or dict()
+    headers = dict(headers) if headers else dict()
     if host: headers["host"] = host
     if authorization: headers["Authorization"] = "Basic %s" % authorization
     url = url + "?" + data if data else url
     url = str(url)
-    file = _resolve(url, name, headers, None, TIMEOUT)
+    file = _resolve(url, name, headers, None, silent, timeout)
     try: result = file.read()
     finally: file.close()
+    info = file.info()
+    location = info.get("Location", None) if redirect else None
+    if location: return _redirect(
+        location,
+        scheme,
+        host,
+        handle = handle,
+        silent = silent,
+        redirect = redirect,
+        timeout = timeout
+    )
     return (result, file) if handle else result
 
 def _method_payload(
@@ -296,21 +430,28 @@ def _method_payload(
     headers = None,
     mime = None,
     handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None,
     **kwargs
 ):
     if handle == None: handle = False
+    if silent == None: silent = False
+    if redirect == None: redirect = False
+    if timeout == None: timeout = TIMEOUT
     values = kwargs or dict()
 
-    url, host, authorization = _parse_url(url)
+    url, scheme, host, authorization, extra = _parse_url(url)
+    if extra: values.update(extra)
     data_e = _urlencode(values)
 
-    if data:
+    if not data == None:
         url = url + "?" + data_e if data_e else url
-    elif data_j:
+    elif not data_j == None:
         data = json.dumps(data_j)
         url = url + "?" + data_e if data_e else url
         mime = mime or "application/json"
-    elif data_m:
+    elif not data_m == None:
         url = url + "?" + data_e if data_e else url
         content_type, data = _encode_multipart(
             data_m, mime = mime, doseq = True
@@ -323,46 +464,115 @@ def _method_payload(
     data = legacy.bytes(data)
     length = len(data) if data else 0
 
-    headers = headers or dict()
+    headers = dict(headers) if headers else dict()
     headers["Content-Length"] = length
     if mime: headers["Content-Type"] = mime
     if host: headers["Host"] = host
     if authorization: headers["Authorization"] = "Basic %s" % authorization
     url = str(url)
 
-    file = _resolve(url, name, headers, data, TIMEOUT)
+    file = _resolve(url, name, headers, data, silent, timeout)
     try: result = file.read()
     finally: file.close()
+
+    info = file.info()
+
+    location = info.get("Location", None) if redirect else None
+    if location: return _redirect(
+        location,
+        scheme,
+        host,
+        handle = handle,
+        silent = silent,
+        redirect = redirect,
+        timeout = timeout
+    )
+
     return (result, file) if handle else result
 
+def _redirect(
+    location,
+    scheme,
+    host,
+    handle = None,
+    silent = None,
+    redirect = None,
+    timeout = None
+):
+    is_relative = location.startswith("/")
+    if is_relative: location = scheme + "://" + host + location
+    return get_json(
+        location,
+        handle = handle,
+        silent = silent,
+        redirect = redirect,
+        timeout = timeout
+    )
+
 def _resolve(*args, **kwargs):
+    # obtains the reference to the global set of variables, so
+    # that it's possible to obtain the proper resolver method
+    # according to the requested client
     _global = globals()
+
+    # tries to retrieve the global configuration values that
+    # will condition the way the request is going to be performed
     client = config.conf("HTTP_CLIENT", "netius")
-    client = kwargs.get("client", client)
-    if "client" in kwargs: del kwargs["client"]
-    resolver = _global.get("_resolve_" + client, _resolve_base)
+
+    # tries to determine the set of configurations requested on
+    # a request basis (not global) these have priority when
+    # compared with the global configuration ones
+    client = kwargs.pop("client", client)
+
+    # tries to retrieve the reference to the resolve method for the
+    # current client and then runs it, retrieve then the final result,
+    # note that the result structure may be engine dependent
+    resolver = _global.get("_resolve_" + client, _resolve_legacy)
     try: result = resolver(*args, **kwargs)
-    except ImportError: result = _resolve_base(*args, **kwargs)
+    except ImportError: result = _resolve_legacy(*args, **kwargs)
     return result
 
-def _resolve_base(url, method, headers, data, timeout):
+def _resolve_legacy(url, method, headers, data, silent, timeout, **kwargs):
     opener = legacy.build_opener(legacy.HTTPHandler)
     request = legacy.Request(url, data = data, headers = headers)
     request.get_method = lambda: method
     return opener.open(request, timeout = timeout)
 
-def _resolve_netius(url, method, headers, data, timeout):
+def _resolve_requests(url, method, headers, data, silent, timeout, **kwargs):
+    import requests
+    method = method.lower()
+    caller = getattr(requests, method)
+    result = caller(url, headers = headers, data = data, timeout = timeout)
+    response = HTTPResponse(
+        data = result.content,
+        code = result.status_code,
+        headers = result.headers
+    )
+    code = response.getcode()
+    is_error = _is_error(code)
+    if is_error: raise legacy.HTTPError(
+        url, code, "HTTP retrieval problem", None, response
+    )
+    return response
+
+def _resolve_netius(url, method, headers, data, silent, timeout, **kwargs):
     import netius.clients
+    silent = silent or False
+    silent |= not common.is_devel()
+    level = logging.CRITICAL if silent else logging.DEBUG
+    level = kwargs.get("level", level)
     result = netius.clients.HTTPClient.method_s(
         method,
         url,
         headers = headers,
         data = data,
-        async = False
+        async = False,
+        timeout = timeout,
+        level = level
     )
     response = netius.clients.HTTPClient.to_response(result)
     code = response.getcode()
-    is_error = code // 100 in (4, 5) if code else True
+    is_error = _is_error(code)
     if is_error: raise legacy.HTTPError(
         url, code, "HTTP retrieval problem", None, response
     )
@@ -370,7 +580,8 @@ def _resolve_netius(url, method, headers, data, timeout):
 
 def _parse_url(url):
     parse = legacy.urlparse(url)
-    secure = parse.scheme == "https"
+    scheme = parse.scheme
+    secure = scheme == "https"
     default = 443 if secure else 80
     port = parse.port or default
     url = parse.scheme + "://" + parse.hostname + ":" + str(port) + parse.path
@@ -378,13 +589,9 @@ def _parse_url(url):
     else: host = parse.hostname + ":" + str(port)
     username = parse.username
     password = parse.password
-    if username and password:
-        payload = "%s:%s" % (username, password)
-        payload = legacy.bytes(payload)
-        authorization = base64.b64encode(payload)
-        authorization = legacy.str(authorization)
-    else: authorization = None
-    return (url, host, authorization)
+    authorization = _authorization(username, password)
+    params = _params(parse.query)
+    return (url, scheme, host, authorization, params)
 
 def _result(data, info = {}, force = False, strict = False):
     # tries to retrieve the content type value from the headers
@@ -408,6 +615,35 @@ def _result(data, info = {}, force = False, strict = False):
         if strict: raise
     return data
 
+def _params(query):
+    # creates the dictionary that is going to be used to store the
+    # complete information regarding the parameters in query
+    params = dict()
+
+    # validates that the provided query value is valid and if
+    # that's not the case returns the created parameters immediately
+    # (empty parameters are returned)
+    if not query: return params
+
+    # splits the query value around the initial parameter separator
+    # symbol and iterates over each of them to parse them and create
+    # the proper parameters dictionary (of lists)
+    query_s = query.split("&")
+    for part in query_s:
+        parts = part.split("=", 1)
+        if len(parts) == 1: value = ""
+        else: value = parts[1]
+        key = parts[0]
+        key = legacy.unquote_plus(key)
+        value = legacy.unquote_plus(value)
+        param = params.get(key, [])
+        param.append(value)
+        params[key] = param
+
+    # returns the final parameters dictionary to the caller method
+    # so that it may be used as a proper structure representation
+    return params
+
 def _urlencode(values, as_string = True):
     # creates the dictionary that will hold the final
     # dictionary of values (without the unset andÂ´
@@ -473,6 +709,15 @@ def _quote(values, plus = False, safe = "/"):
 
     return final
 
+def _authorization(username, password):
+    if not username: return None
+    if not password: return None
+    payload = "%s:%s" % (username, password)
+    payload = legacy.bytes(payload)
+    authorization = base64.b64encode(payload)
+    authorization = legacy.str(authorization)
+    return authorization
+
 def _encode_multipart(fields, mime = None, doseq = False):
     mime = mime or "multipart/form-data"
     boundary = _create_boundary(fields, doseq = doseq)
@@ -485,9 +730,8 @@ def _encode_multipart(fields, mime = None, doseq = False):
         values = values if is_list else [values]
 
         for value in values:
-            value_t = type(value)
 
-            if value_t == dict:
+            if isinstance(value, dict):
                 header_l = []
                 data = None
                 for key, item in value.items():
@@ -495,10 +739,14 @@ def _encode_multipart(fields, mime = None, doseq = False):
                     else: header_l.append("%s: %s" % (key, item))
                 value = data
                 header = "\r\n".join(header_l)
-            elif value_t == tuple:
+            elif isinstance(value, tuple):
+                content_type = None
+                if len(value) == 2: name, contents = value
+                else: name, content_type, contents = value
                 header = "Content-Disposition: form-data; name=\"%s\"; filename=\"%s\"" %\
-                    (key, value[0])
-                value = value[1]
+                    (key, name)
+                if content_type: header += "\r\nContent-Type: %s" % content_type
+                value = contents
             else:
                 header = "Content-Disposition: form-data; name=\"%s\"" % key
                 value = _encode(value)
@@ -535,10 +783,8 @@ def _try_boundary(fields, boundary, doseq = False):
         values = values if is_list else [values]
 
         for value in values:
-            value_t = type(value)
-
-            if value_t == dict: name = ""; value = value.get("data", b"")
-            elif value_t == tuple: name = value[0]; value = value[1]
+            if isinstance(value, dict): name = ""; value = value.get("data", b"")
+            elif isinstance(value, tuple): name = value[0] or ""; value = value[1]
             else: name = ""; value = _encode(value)
 
             if not key.find(boundary) == -1: return False
@@ -547,8 +793,39 @@ def _try_boundary(fields, boundary, doseq = False):
 
     return True
 
+def _is_error(code):
+    return code // 100 in (4, 5) if code else True
+
 def _encode(value, encoding = "utf-8"):
     value_t = type(value)
     if value_t == legacy.BYTES: return value
     elif value_t == legacy.UNICODE: return value.encode(encoding)
     return legacy.bytes(str(value))
+
+class HTTPResponse(object):
+    """
+    Compatibility object to be used by HTTP libraries that do
+    not support the legacy HTTP response object as a return
+    for any of their structures.
+    """
+
+    def __init__(self, data = None, code = 200, status = None, headers = None):
+        self.data = data
+        self.code = code
+        self.status = status
+        self.headers = headers
+
+    def read(self):
+        return self.data
+
+    def readline(self):
+        return self.read()
+
+    def close(self):
+        pass
+
+    def getcode(self):
+        return self.code
+
+    def info(self):
+        return self.headers
